import logging
import sys
from datetime import datetime
from typing import List

import typer
from tabulate import tabulate

from pythonanywhere.scripts_commons import get_logger, get_task_from_id, tabulate_formats
from pythonanywhere.snakesay import snakesay
from pythonanywhere.task import Task, TaskList

app = typer.Typer()


@app.command()
def set(
    command: str = typer.Option(
        ..., "-c", "--command", help="Task's command to be scheduled"
    ),
    hour: int = typer.Option(
        None,
        "-h",
        "--hour",
        min=0,
        max=23,
        help="Sets the task to be performed daily at HOUR",
    ),
    minute: int = typer.Option(
        ...,
        "-m",
        "--minute",
        min=0,
        max=59,
        help="Minute on which the task will be executed",
    ),
    disabled: bool = typer.Option(
        False, "-d", "--disabled", help="Creates disabled task (otherwise enabled)"
    ),
):
    """Create a scheduled task.

    Two categories of tasks are available: daily and hourly.
    Both kinds require a command to run and scheduled time. In order to create a
    daily task provide hour and minute; to create hourly task provide only minute.
    If task is intended to be enabled later add --disabled flag.

    Example:
      Create a daily task to be run at 13:15:

        pa schedule set --command "echo foo" --hour 13 --minute 15

      Create an inactive hourly task to be run 27 minutes past every hour:

        pa schedule set --command "echo bar" --minute 27 --disabled

    Note:
      Once task is created its behavior may be altered later on with
      `pa schedule update` or deleted with `pa schedule delete`
      commands."""

    logger = get_logger(set_info=True)

    task = Task.to_be_created(
        command=command, hour=hour, minute=minute, disabled=disabled
    )
    try:
        task.create_schedule()
    except Exception as e:
        logger.warning(snakesay(str(e)))


delete_app = typer.Typer()
app.add_typer(
    delete_app, name="delete", help="Delete scheduled task(s) by id or nuke'em all."
)


@delete_app.command("all", help="Delete all scheduled tasks.")
def delete_all_tasks(
    force: bool = typer.Option(
        False, "-f", "--force", help="Turns off user confirmation before deleting tasks"
    ),
):
    get_logger(set_info=True)

    if not force:
        user_response = typer.confirm(
            "This will irrevocably delete all your tasks, proceed?"
        )
        if not user_response:
            return None

    for task in TaskList().tasks:
        task.delete_schedule()


@delete_app.command(
    "id",
    help="""\b
    Delete one or more scheduled tasks by id.
    ID_NUMBERS may be acquired with `pa schedule list`
    """,
)
def delete_task_by_id(id_numbers: List[int] = typer.Argument(...)):
    get_logger(set_info=True)

    for task_id in id_numbers:
        task = get_task_from_id(task_id, no_exit=True)
        task.delete_schedule()


@app.command()
def get(
    task_id: int = typer.Argument(..., metavar="id"),
    command: bool = typer.Option(
        False, "-c", "--command", help="Prints task's command"
    ),
    enabled: bool = typer.Option(
        False, "-e", "--enabled", help="Prints task's enabled status (True or False)"
    ),
    expiry: bool = typer.Option(
        False, "-x", "--expiry", help="Prints task's expiry date"
    ),
    minute: bool = typer.Option(
        False, "-m", "--minute", help="Prints task's scheduled minute"
    ),
    hour: bool = typer.Option(
        False, "-o", "--hour", help="Prints task's scheduled hour (if daily)"
    ),
    interval: bool = typer.Option(
        False, "-i", "--interval", help="Prints task's frequency (daily or hourly)"
    ),
    logfile: bool = typer.Option(
        False, "-l", "--logfile", help="Prints task's current log file path"
    ),
    printable_time: bool = typer.Option(
        False, "-p", "--printable-time", help="Prints task's scheduled time"
    ),
    no_spec: bool = typer.Option(
        False, "-n", "--no-spec", help="Prints only values without spec names"
    ),
    snake: bool = typer.Option(
        False, "-s", "--snakesay", help="Turns on snakesay... because why not"
    )
):
    """Get scheduled task's specs.

    Available specs are: command, enabled, interval, hour, minute, printable-time,
    logfile, expiry. If no option specified, script will output all mentioned specs.

    Note that logfile query provides path for current (last) logfile. There may be
    several logfiles for each task.
    If task has been updated (e.g. by `pa_update_scheduled_task.py` script) logfile
    name has been changed too, but the file will not be created until first execution
    of the task. Thus getting logfile path via API call does not necessarily mean the
    file exists on the server yet.

    Note:
    Task ID may be found using pa schedule list command.

    Example:
    Get all specs for task with id 42:

        pa schedule get 42

    Get only logfile name for task with id 42:

        pa schedule get 42 --logfile --no-spec"""

    kwargs = {k: v for k, v in locals().items() if k != "task_id"}
    logger = get_logger(set_info=True)

    task = get_task_from_id(task_id)

    print_snake = kwargs.pop("snake")
    print_only_values = kwargs.pop("no_spec")

    specs = (
        {spec: getattr(task, spec) for spec in kwargs if kwargs[spec]}
        if any([val for val in kwargs.values()])
        else {spec: getattr(task, spec) for spec in kwargs}
    )

    if specs.get("logfile"):
        specs.update({"logfile": task.logfile.replace(f"/user/{task.user}/files", "")})

    intro = f"Task {task_id} specs: "
    if print_only_values:
        specs = "\n".join([str(val) for val in specs.values()])
        logger.info(specs)
    elif print_snake:
        specs = [f"<{spec}>: {value}" for spec, value in specs.items()]
        specs.sort()
        logger.info(snakesay(intro + ", ".join(specs)))
    else:
        table = [[spec, val] for spec, val in specs.items()]
        table.sort(key=lambda x: x[0])
        logger.info(intro)
        logger.info(tabulate(table, tablefmt="simple"))


def tablefmt_callback(value: str):
    if value not in tabulate_formats:
        raise typer.BadParameter(f"Table format has to be one of: {', '.join(tabulate_formats)}")
    return value


@app.command("list")
def list_(
    tablefmt: str = typer.Option(
        "simple", "-f", "--format", help="Table format", callback=tablefmt_callback
    )
):
    """Get list of user's scheduled tasks as a table with columns:
    id, interval, at (hour:minute/minute past), status (enabled/disabled), command.

    Note:
    This script provides an overview of all tasks. Once a task id is
    known and some specific data is required it's more convenient to get
    it using `pa schedule get` command instead of parsing the table.
    """

    logger = get_logger(set_info=True)

    headers = "id", "interval", "at", "status", "command"
    attrs = "task_id", "interval", "printable_time", "enabled", "command"

    def stringify_values(task, attr):
        value = getattr(task, attr)
        if attr == "enabled":
            value = "enabled" if value else "disabled"
        return value

    table = [[stringify_values(task, attr) for attr in attrs] for task in TaskList().tasks]
    msg = tabulate(table, headers, tablefmt=tablefmt) if table else snakesay("No scheduled tasks")
    logger.info(msg)


@app.command()
def update(
    task_id: int = typer.Argument(..., metavar="id"),
    command: str = typer.Option(
        None,
        "-c",
        "--command",
        help="Changes command to COMMAND (multiword commands should be quoted)"
    ),
    hour: int = typer.Option(
        None,
        "-o",
        "--hour",
        min=0,
        max=23,
        help="Changes hour to HOUR (in 24h format)"
    ),
    minute: int = typer.Option(
        None,
        "-m",
        "--minute",
        min=0,
        max=59,
        help="Changes minute to MINUTE"
    ),
    disable: bool = typer.Option(False, "-d", "--disable", help="Disables task"),
    enable: bool = typer.Option(False, "-e", "--enable", help="Enables task"),
    toggle_enabled: bool = typer.Option(
        False, "-t", "--toggle-enabled", help="Toggles enable/disable state"
    ),
    daily: bool = typer.Option(
        False,
        "-a",
        "--daily",
        help=(
            "Switches interval to daily "
            "(when --hour is not provided, sets it automatically to current hour)"
        )
    ),
    hourly: bool = typer.Option(
        False,
        "-u",
        "--hourly",
        help="Switches interval to hourly (takes precedence over --hour, i.e. sets hour to None)"
    ),
    quiet: bool = typer.Option(False, "-q", "--quiet", help="Turns off messages"),
    porcelain: bool = typer.Option(
        False, "-p", "--porcelain", help="Prints message in easy-to-parse format"
    ),
):
    """Update a scheduled task.

    Note that logfile name will change after updating the task but it won't be
    created until first execution of the task.
    To change interval from hourly to daily use --daily flag and provide --hour.
    When --daily flag is not accompanied with --hour, new hour for the task
    will be automatically set to current hour.
    When changing interval from daily to hourly --hour flag is ignored.

    Example:
    Change command for a scheduled task 42:

        pa schedule update 42 --command "echo new command"

    Change interval of the task 42 from hourly to daily to be run at 10 am:

        pa schedule update 42 --hour 10

    Change interval of the task 42 from daily to hourly and set new minute:

        pa schedule update 42 --minute 13 --hourly"""

    kwargs = {k: v for k, v in locals().items() if k != "task_id"}
    logger = get_logger()

    porcelain = kwargs.pop("porcelain")
    if not kwargs.pop("quiet"):
        logger.setLevel(logging.INFO)

    if not any(kwargs.values()):
        msg = "Nothing to update!"
        logger.warning(msg if porcelain else snakesay(msg))
        sys.exit(1)

    if kwargs.pop("hourly"):
        kwargs["interval"] = "hourly"
    if kwargs.pop("daily"):
        kwargs["hour"] = kwargs["hour"] if kwargs["hour"] else datetime.now().hour
        kwargs["interval"] = "daily"

    task = get_task_from_id(task_id)

    enable_opt = [k for k in ["toggle_enabled", "disable", "enable"] if kwargs.pop(k)]
    params = {k: v for k, v in kwargs.items() if v}
    if enable_opt:
        lookup = {"toggle_enabled": not task.enabled, "disable": False, "enable": True}
        params.update({"enabled": lookup[enable_opt[0]]})

    try:
        task.update_schedule(params, porcelain=porcelain)
    except Exception as e:
        logger.warning(snakesay(str(e)))



import getpass
import os
from pathlib import Path
from textwrap import dedent

from dateutil.parser import parse

from pythonanywhere.api.base import PYTHON_VERSIONS, call_api, get_api_endpoint
from pythonanywhere.exceptions import SanityException
from pythonanywhere.snakesay import snakesay


class Webapp:
    def __init__(self, domain):
        self.domain = domain

    def __eq__(self, other):
        return self.domain == other.domain

    def sanity_checks(self, nuke):
        print(snakesay("Running API sanity checks"))
        token = os.environ.get("API_TOKEN")
        if not token:
            raise SanityException(
                dedent(
                    """
                Could not find your API token.
                You may need to create it on the Accounts page?
                You will also need to close this console and open a new one once you've done that.
                """
                )
            )

        if nuke:
            return

        url = get_api_endpoint().format(username=getpass.getuser(), flavor="webapps") + self.domain + "/"
        response = call_api(url, "get")
        if response.status_code == 200:
            raise SanityException(
                "You already have a webapp for {domain}.\n\nUse the --nuke option if you want to replace it.".format(
                    domain=self.domain
                )
            )

    def create(self, python_version, virtualenv_path, project_path, nuke):
        print(snakesay("Creating web app via API"))
        if nuke:
            webapp_url = get_api_endpoint().format(username=getpass.getuser(), flavor="webapps") + self.domain + "/"
            call_api(webapp_url, "delete")
        post_url = get_api_endpoint().format(username=getpass.getuser(), flavor="webapps")
        patch_url = post_url + self.domain + "/"
        response = call_api(
            post_url, "post", data={"domain_name": self.domain, "python_version": PYTHON_VERSIONS[python_version]}
        )
        if not response.ok or response.json().get("status") == "ERROR":
            raise Exception(
                f"POST to create webapp via API failed, got {response}:{response.text}"
            )
        response = call_api(
            patch_url, "patch", data={"virtualenv_path": virtualenv_path, "source_directory": project_path}
        )
        if not response.ok:
            raise Exception(
                "PATCH to set virtualenv path and source directory via API failed,"
                "got {response}:{response_text}".format(response=response, response_text=response.text)
            )

    def add_default_static_files_mappings(self, project_path):
        print(snakesay("Adding static files mappings for /static/ and /media/"))

        url = (
            get_api_endpoint().format(username=getpass.getuser(), flavor="webapps") + self.domain + "/static_files/"
        )
        call_api(url, "post", json=dict(url="/static/", path=str(Path(project_path) / "static")))
        call_api(url, "post", json=dict(url="/media/", path=str(Path(project_path) / "media")))

    def reload(self):
        print(snakesay(f"Reloading {self.domain} via API"))
        url = get_api_endpoint().format(username=getpass.getuser(), flavor="webapps") + self.domain + "/reload/"
        response = call_api(url, "post")
        if not response.ok:
            if response.status_code == 409 and response.json()["error"] == "cname_error":
                print(
                    snakesay(
                        dedent("""
                            Could not find a CNAME for your website.  If you're using an A record,
                            CloudFlare, or some other way of pointing your domain at PythonAnywhere
                            then that should not be a problem.  If you're not, you should double-check
                            your DNS setup.
                        """)
                    )
                )
                return
            raise Exception(
                f"POST to reload webapp via API failed, got {response}:{response.text}"
            )

    def set_ssl(self, certificate, private_key):
        print(snakesay(f"Setting up SSL for {self.domain} via API"))
        url = get_api_endpoint().format(username=getpass.getuser(), flavor="webapps") + self.domain + "/ssl/"
        response = call_api(url, "post", json={"cert": certificate, "private_key": private_key})
        if not response.ok:
            raise Exception(
                dedent(
                    """
                    POST to set SSL details via API failed, got {response}:{response_text}
                    If you just created an API token, you need to set the API_TOKEN environment variable or start a
                    new console.  Also you need to have setup a `{domain}` PythonAnywhere webapp for this to work.
                    """
                ).format(response=response, response_text=response.text, domain=self.domain)
            )

    def get_ssl_info(self):
        url = get_api_endpoint().format(username=getpass.getuser(), flavor="webapps") + self.domain + "/ssl/"
        response = call_api(url, "get")
        if not response.ok:
            raise Exception(
                f"GET SSL details via API failed, got {response}:{response.text}"
            )

        result = response.json()
        result["not_after"] = parse(result["not_after"])
        return result

    def delete_log(self, log_type, index=0):
        if index:
            print(
                snakesay(
                    "Deleting old (archive number {index}) {type} log file for {domain} via API".format(
                        index=index, type=log_type, domain=self.domain
                    )
                )
            )
        else:
            print(
                snakesay(
                    f"Deleting current {log_type} log file for {self.domain} via API"
                )
            )

        if index == 1:
            url = get_api_endpoint().format(
                username=getpass.getuser(), flavor="files"
            ) + f"path/var/log/{self.domain}.{log_type}.log.1/"
        elif index > 1:
            url = get_api_endpoint().format(
                username=getpass.getuser(), flavor="files"
            ) + f"path/var/log/{self.domain}.{log_type}.log.{index}.gz/"
        else:
            url = get_api_endpoint().format(
                username=getpass.getuser(), flavor="files"
            ) + f"path/var/log/{self.domain}.{log_type}.log/"
        response = call_api(url, "delete")
        if not response.ok:
            raise Exception(
                f"DELETE log file via API failed, got {response}:{response.text}"
            )

    def get_log_info(self):
        url = get_api_endpoint().format(username=getpass.getuser(), flavor="files") + "tree/?path=/var/log/"
        response = call_api(url, "get")
        if not response.ok:
            raise Exception(
                f"GET log files info via API failed, got {response}:{response.text}"
            )
        file_list = response.json()
        log_types = ["access", "error", "server"]
        logs = {"access": [], "error": [], "server": []}
        log_prefix = f"/var/log/{self.domain}."
        for file_name in file_list:
            if type(file_name) == str and file_name.startswith(log_prefix):
                log = file_name[len(log_prefix):].split(".")
                if log[0] in log_types:
                    log_type = log[0]
                    if log[-1] == "log":
                        log_index = 0
                    elif log[-1] == "1":
                        log_index = 1
                    elif log[-1] == "gz":
                        log_index = int(log[-2])
                    else:
                        continue
                    logs[log_type].append(log_index)
        return logs




import typer

from cli import django
from cli import schedule
from cli import webapp
from cli import path

help = """This is a new experimental PythonAnywhere cli client.

It was build with typer & click under the hood.
"""

app = typer.Typer(help=help)
app.add_typer(django.app, name="django", help="Makes Django Girls tutorial projects deployment easy")
app.add_typer(schedule.app, name="schedule", help="Manage scheduled tasks")
app.add_typer(webapp.app, name="webapp", help="Everything for web apps")
app.add_typer(path.app, name="path", help="Perform some operations on files")


if __name__ == "__main__":
    app(prog_name="pa")
